-------     =================================================================
------------  programmes d'essai pour l'exclusion mutuelle
-------     =================================================================

with Ada.Text_IO; use Ada.Text_IO;
with Global; use Global;
-- with Ada.Calendar; use Ada.Calendar;

with No_Control, Simpleton, Dekker_2, Peterson_2, Lamport_N, Semaphores;

procedure Mutual_Exclusion is
	-- number of processes competing for mutual exclusion
	-- Global.Nb_Proc : Constant  number of processes
	-- type Global.Process_Id is mod Nb_Proc; -- process unique names

	Next_Run : Global.Algorithm := Algorithm'First; -- starts with No_Control
   --------------
   procedure Yield is
   -- used to simulate process switching by processor allocator
   begin
	delay(0.1*duration(Global.Next_Hazard)) ; -- Next_Hazard in the interval 0.0..1.0
	null;
   end Yield;
   ---------------
   package Count is
	procedure Update(Amount : in Integer; Actor : in Process_Id);
	procedure Reset;
   end Count;
   package body Count is
	Client_Count : Integer := 0;
	procedure Update(Amount : in Integer; Actor : in Process_Id) is
		X : Integer;
	begin
		-- the following is a critical section to be executed in mutual exclusion
		X := Client_Count; Yield;	-- P1
		X := X + Amount; Yield;		-- P2
		Client_Count := X; Yield;	-- P3
		Put_Line(Integer'Image(Client_Count) & "      after client "
					& Process_Id'Image(Actor) & "  adding " & Integer'Image(Amount));
		Yield;	
	end Update;
	procedure Reset is begin Client_Count:= 0; end Reset;
   end Count;
   ---------------
  --  package Dummy_Critical_Section is
	-- procedure Enter(Requestor : in Global.Process_Id);
	-- procedure Quit(Requestor : in Global.Process_Id);
  -- end Dummy_Critical_Section;
   -- package body Dummy_Critical_Section is
	-- initialisation protocol
	-- procedure Enter(Requestor : in Global.Process_Id) is begin null; end Enter;
	-- procedure Quit(Requestor : in Global.Process_Id) is begin null; end Quit;
   -- begin
    -- initialisation protocol continued if necessary
	-- null;
   -- end Dummy_Critical_Section;
   
   type Access_Control is access procedure(Requestor : in Global.Process_Id);
   Enter : Access_Control;
   Quit : Access_Control;
   ---------------
   package Unique_Name is
	function Get return Process_Id;
   end Unique_Name;
   package body Unique_Name is
	Current_Name : Process_Id := Process_Id'Last;
	function Get return Process_Id is 
	begin 
		Current_Name := Current_Name + 1; 
		return Current_Name; 
	end Get;
   end Unique_Name;
   ---------------
   protected Termination is 
		procedure Signal;
		entry Wait;
   private
		Count: Integer := 0;
   end Termination;
   protected body Termination is 
		procedure Signal is begin Count := Count + 1; end Signal;
		entry Wait when Count = 2 is
		begin Count := 0; end Wait;
   end Termination;

   ---------------
   task type A_Process(My_Id : Process_Id := Unique_Name.Get);
   task body A_Process is
      Value : Integer := 1;
	  -- My_Id : Process_Id 
	begin
      for I in 1 .. 16 loop

		Enter(My_Id);
		Count.update (Value, My_Id);
		Quit(My_Id);

	  end loop;
	  Termination.Signal;
   end A_Process;
   ----------------

begin
      
   loop
		Put_Line ("##### mutual exclusion algorithm " & Algorithm'Image(Next_Run));
		Count.Reset; -- starts a new run with Client_Count value at zero

		---------------
		case Next_Run is
			when Global.No_Control		=> 
				Enter := No_Control.Enter'Access; Quit := No_Control.Quit'Access;
			when Global.Simpleton		=>  
				Enter := Simpleton.Enter'Access; Quit := Simpleton.Quit'Access;
			when Global.Dekker_2		=> 
				Enter := Dekker_2.Enter'Access; Quit := Dekker_2.Quit'Access;
			when Global.Peterson_2		=>  
				Enter := Peterson_2.Enter'Access; Quit := Peterson_2.Quit'Access;
			when Global.Lamport_N		=>  
				Enter := Lamport_N.Enter'Access; Quit := Lamport_N.Quit'Access;
			when Global.Semaphores		=>  
				Enter := Semaphores.Enter'Access; Quit := Semaphores.Quit'Access;
			when others					=>  
				Enter := No_Control.Enter'Access; Quit := No_Control.Quit'Access;
		end case;
		---------------

		-- Create processes
		declare
			PP : A_Process;
			QQ : A_Process;
		begin null; end ;
		Termination.Wait;
		exit when Next_Run = Global.Algorithm'Last; 
		Next_Run := Global.Algorithm'Succ(Next_Run);
	end loop;
   Put_Line("##### end of experimentation");
   
end Mutual_Exclusion;
